@@ -166,7 +166,120 @@ class Softmax(Layer):
             dA[index] = np.dot(jacobian_matrix,
             single_dvalues)
         return dA
+
+#%%LiveStats
+class LiveStats():
+    def __init__(self, type=-1, f1=0, f2=1, row_select="limited"):
+        super().__init__()
+        self.type = type
+        self.f1 = f1
+        self.f2 = f2
+        self.row_select = row_select
     
+    def run(self, model, figure):
+        import matplotlib.pyplot as plt
+        if self.type == 1:
+            figure.clear()
+            
+            plt.subplot(1,2,1)
+            plt.title("Loss Evolution")
+            plt.xlabel("Epoch")
+            plt.ylabel("Loss")
+            plt.plot(model.train_loss_records, color="#1f77b4", label=f"training dataset: {round(float(model.train_loss_records[-1]), 3)}")
+            if model.val_inputs is not None and model.val_targets is not None:
+                plt.plot(model.val_loss_records, color="orange", label=f"validation dataset: {round(float(model.val_loss_records[-1]), 3)}")
+            plt.legend()
+            
+            plt.subplot(1,2,2)
+            plt.title("Accuracy Evolution")
+            plt.xlabel("Epoch")
+            plt.ylabel("Accuracy")
+            plt.plot(model.train_accuracy_records, color="#1f77b4", label=f"training dataset: {round(float(model.train_accuracy_records[-1]), 3)}")
+            if model.val_inputs is not None and model.val_targets is not None:
+                plt.plot(model.val_accuracy_records, color="orange", label=f"validation dataset: {round(float(model.val_accuracy_records[-1]), 3)}")
+            plt.legend()
+            
+            figure.canvas.draw()
+            figure.canvas.flush_events()
+            
+        elif self.type == 2:
+            figure.clear()
+            
+            plt.subplot(1,1,1)
+            if len(model.train_inputs[0,:]) <= 2 and len(model.train_targets[0,:]) == 1:
+                if len(model.train_inputs[:,0]) > 1000 and self.row_select=="limited":
+                    X_set, y_set = model.train_inputs[:1000,:], model.train_targets[:1000,:]
+                elif len(model.train_inputs[:,0]) > 1000 and self.row_select=="full":
+                    X_set, y_set = model.train_inputs, model.train_targets
+                elif (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="full") or (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="limited"):
+                    X_set, y_set = model.train_inputs, model.train_targets
+                else:
+                    raise ValueError("invalid value for 'row_select'")
+                    
+                X1, X2 = np.meshgrid(np.arange(start = X_set[:, self.f1].min() - 1, stop = X_set[:, self.f1].max() + 1, step = 0.01),
+                                     np.arange(start = X_set[:, self.f2].min() - 1, stop = X_set[:, self.f2].max() + 1, step = 0.01))
+                plt.contourf(X1, X2, model.predict(np.array([X1.ravel(), X2.ravel()]).T).reshape(X1.shape),
+                             alpha = 0.3, cmap = "coolwarm")
+            
+            if len(model.train_inputs[:,0]) > 1000 and self.row_select=="limited":
+                plt.scatter(model.train_inputs[:1000,self.f1], model.train_inputs[:1000,self.f2], c=model.train_outputs[:1000,:], cmap="coolwarm", alpha=1)
+            elif len(model.train_inputs[:,0]) > 1000 and self.row_select=="full":
+                plt.scatter(model.train_inputs[:,self.f1], model.train_inputs[:,self.f2], c=model.train_outputs[:,:], cmap="coolwarm", alpha=1)
+            elif (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="full") or (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="limited"):
+                plt.scatter(model.train_inputs[:,self.f1], model.train_inputs[:,self.f2], c=model.train_outputs[:,:], cmap="coolwarm", alpha=1)
+            else:
+                raise ValueError("invalid value for 'row_select'")
+                
+            figure.canvas.draw()
+            figure.canvas.flush_events()
+            
+        elif self.type == -1:
+            figure.clear()
+            
+            plt.subplot(1, 2, 2)
+            if len(model.train_inputs[0,:]) <= 2 and len(model.train_targets[0,:]) == 1:
+                if len(model.train_inputs[:,0]) > 1000 and self.row_select=="limited":
+                    X_set, y_set = model.train_inputs[:1000,:], model.train_targets[:1000,:]
+                elif len(model.train_inputs[:,0]) > 1000 and self.row_select=="full":
+                    X_set, y_set = model.train_inputs, model.train_targets
+                elif (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="full") or (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="limited"):
+                    X_set, y_set = model.train_inputs, model.train_targets
+                else:
+                    raise ValueError("invalid value for 'row_select'")
+                X1, X2 = np.meshgrid(np.arange(start = X_set[:, self.f1].min() - 1, stop = X_set[:, self.f1].max() + 1, step = 0.01),
+                                     np.arange(start = X_set[:, self.f2].min() - 1, stop = X_set[:, self.f2].max() + 1, step = 0.01))
+                plt.contourf(X1, X2, model.predict(np.array([X1.ravel(), X2.ravel()]).T).reshape(X1.shape),
+                             alpha = 0.3, cmap = "coolwarm")
+                
+            if len(model.train_inputs[:,0]) > 1000 and self.row_select=="limited":
+                plt.scatter(model.train_inputs[:1000,self.f1], model.train_inputs[:1000,self.f2], c=model.train_outputs[:1000,:], cmap="coolwarm", alpha=1)
+            elif len(model.train_inputs[:,0]) > 1000 and self.row_select=="full":
+                plt.scatter(model.train_inputs[:,self.f1], model.train_inputs[:,self.f2], c=model.train_outputs[:,:], cmap="coolwarm", alpha=1)
+            elif (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="full") or (len(model.train_inputs[:,0]) <= 1000 and self.row_select=="limited"):
+                plt.scatter(model.train_inputs[:,self.f1], model.train_inputs[:,self.f2], c=model.train_outputs[:,:], cmap="coolwarm", alpha=1)
+            else:
+                raise ValueError("invalid value for 'row_select'")
+            
+            plt.subplot(2,2,1)
+            plt.title("Loss Evolution")
+            plt.ylabel("Loss")
+            plt.plot(model.train_loss_records, color="#1f77b4", label=f"training dataset: {round(float(model.train_loss_records[-1]), 3)}")
+            if model.val_inputs is not None and model.val_targets is not None:
+                plt.plot(model.val_loss_records, color="orange", label=f"validation dataset: {round(float(model.val_loss_records[-1]), 3)}")
+            plt.legend()
+            
+            plt.subplot(2,2,3)
+            plt.title("Accuracy Evolution")
+            plt.xlabel("Epoch")
+            plt.ylabel("Accuracy")
+            plt.plot(model.train_accuracy_records, color="#1f77b4", label=f"training dataset: {round(float(model.train_accuracy_records[-1]), 3)}")
+            if model.val_inputs is not None and model.val_targets is not None:
+                plt.plot(model.val_accuracy_records, color="orange", label=f"validation dataset: {round(float(model.val_accuracy_records[-1]), 3)}")
+            plt.legend()
+            
+            figure.canvas.draw()
+            figure.canvas.flush_events()
+            
 #%%Sequential 
 class Sequential():
     def __init__(self, train_inputs, train_targets, val_inputs = None, val_targets = None):
@@ -293,126 +406,18 @@ class Sequential():
         else:
             raise ValueError("unknown value for verbose")
     
-    def live_stats(self,ls_type, figure, F1=0, F2=1, n_row="limited"):
-        import matplotlib.pyplot as plt
-        if ls_type == 1:
-            figure.clear()
+    def fit(self, cost_fn, epochs = 1000, lr = 0.1, momentum = None, verbose = 1, live_stats = None):
+        if live_stats == None:
+            live_stats = LiveStats()
             
-            plt.subplot(1,2,1)
-            plt.title("Loss Evolution")
-            plt.xlabel("Epoch")
-            plt.ylabel("Loss")
-            plt.plot(self.train_loss_records, color="#1f77b4", label=f"training dataset: {round(float(self.train_loss_records[-1]), 3)}")
-            if self.val_inputs is not None and self.val_targets is not None:
-                plt.plot(self.val_loss_records, color="orange", label=f"validation dataset: {round(float(self.val_loss_records[-1]), 3)}")
-            plt.legend()
-            
-            plt.subplot(1,2,2)
-            plt.title("Accuracy Evolution")
-            plt.xlabel("Epoch")
-            plt.ylabel("Accuracy")
-            plt.plot(self.train_accuracy_records, color="#1f77b4", label=f"training dataset: {round(float(self.train_accuracy_records[-1]), 3)}")
-            if self.val_inputs is not None and self.val_targets is not None:
-                plt.plot(self.val_accuracy_records, color="orange", label=f"validation dataset: {round(float(self.val_accuracy_records[-1]), 3)}")
-            plt.legend()
-            
-            figure.canvas.draw()
-            figure.canvas.flush_events()
-            
-        elif ls_type == 2:
-            figure.clear()
-            
-            plt.subplot(1,1,1)
-            if len(self.train_inputs[0,:]) <= 2 and len(self.train_targets[0,:]) == 1:
-                if len(self.train_inputs[:,0]) > 1000 and n_row=="limited":
-                    X_set, y_set = self.train_inputs[:1000,:], self.train_targets[:1000,:]
-                elif len(self.train_inputs[:,0]) > 1000 and n_row=="full":
-                    X_set, y_set = self.train_inputs, self.train_targets
-                elif (len(self.train_inputs[:,0]) <= 1000 and n_row=="full") or (len(self.train_inputs[:,0]) <= 1000 and n_row=="limited"):
-                    X_set, y_set = self.train_inputs, self.train_targets
-                else:
-                    raise ValueError("invalid value for 'n_row'")
-                    
-                X1, X2 = np.meshgrid(np.arange(start = X_set[:, F1].min() - 1, stop = X_set[:, F1].max() + 1, step = 0.01),
-                                     np.arange(start = X_set[:, F2].min() - 1, stop = X_set[:, F2].max() + 1, step = 0.01))
-                plt.contourf(X1, X2, self.predict(np.array([X1.ravel(), X2.ravel()]).T).reshape(X1.shape),
-                             alpha = 0.3, cmap = "coolwarm")
-            
-            if len(self.train_inputs[:,0]) > 1000 and n_row=="limited":
-                plt.scatter(self.train_inputs[:1000,F1], self.train_inputs[:1000,F2], c=self.train_outputs[:1000,:], cmap="coolwarm", alpha=1)
-            elif len(self.train_inputs[:,0]) > 1000 and n_row=="full":
-                plt.scatter(self.train_inputs[:,F1], self.train_inputs[:,F2], c=self.train_outputs[:,:], cmap="coolwarm", alpha=1)
-            elif (len(self.train_inputs[:,0]) <= 1000 and n_row=="full") or (len(self.train_inputs[:,0]) <= 1000 and n_row=="limited"):
-                plt.scatter(self.train_inputs[:,F1], self.train_inputs[:,F2], c=self.train_outputs[:,:], cmap="coolwarm", alpha=1)
-            else:
-                raise ValueError("invalid value for 'n_row'")
-                
-            figure.canvas.draw()
-            figure.canvas.flush_events()
-            
-        elif ls_type == -1:
-            figure.clear()
-            
-            plt.subplot(1, 2, 2)
-            if len(self.train_inputs[0,:]) <= 2 and len(self.train_targets[0,:]) == 1:
-                if len(self.train_inputs[:,0]) > 1000 and n_row=="limited":
-                    X_set, y_set = self.train_inputs[:1000,:], self.train_targets[:1000,:]
-                elif len(self.train_inputs[:,0]) > 1000 and n_row=="full":
-                    X_set, y_set = self.train_inputs, self.train_targets
-                elif (len(self.train_inputs[:,0]) <= 1000 and n_row=="full") or (len(self.train_inputs[:,0]) <= 1000 and n_row=="limited"):
-                    X_set, y_set = self.train_inputs, self.train_targets
-                else:
-                    raise ValueError("invalid value for 'n_row'")
-                X1, X2 = np.meshgrid(np.arange(start = X_set[:, F1].min() - 1, stop = X_set[:, F1].max() + 1, step = 0.01),
-                                     np.arange(start = X_set[:, F2].min() - 1, stop = X_set[:, F2].max() + 1, step = 0.01))
-                plt.contourf(X1, X2, self.predict(np.array([X1.ravel(), X2.ravel()]).T).reshape(X1.shape),
-                             alpha = 0.3, cmap = "coolwarm")
-                
-            if len(self.train_inputs[:,0]) > 1000 and n_row=="limited":
-                plt.scatter(self.train_inputs[:1000,F1], self.train_inputs[:1000,F2], c=self.train_outputs[:1000,:], cmap="coolwarm", alpha=1)
-            elif len(self.train_inputs[:,0]) > 1000 and n_row=="full":
-                plt.scatter(self.train_inputs[:,F1], self.train_inputs[:,F2], c=self.train_outputs[:,:], cmap="coolwarm", alpha=1)
-            elif (len(self.train_inputs[:,0]) <= 1000 and n_row=="full") or (len(self.train_inputs[:,0]) <= 1000 and n_row=="limited"):
-                plt.scatter(self.train_inputs[:,F1], self.train_inputs[:,F2], c=self.train_outputs[:,:], cmap="coolwarm", alpha=1)
-            else:
-                raise ValueError("invalid value for 'n_row'")
-            
-            plt.subplot(2,2,1)
-            plt.title("Loss Evolution")
-            plt.ylabel("Loss")
-            plt.plot(self.train_loss_records, color="#1f77b4", label=f"training dataset: {round(float(self.train_loss_records[-1]), 3)}")
-            if self.val_inputs is not None and self.val_targets is not None:
-                plt.plot(self.val_loss_records, color="orange", label=f"validation dataset: {round(float(self.val_loss_records[-1]), 3)}")
-            plt.legend()
-            
-            plt.subplot(2,2,3)
-            plt.title("Accuracy Evolution")
-            plt.xlabel("Epoch")
-            plt.ylabel("Accuracy")
-            plt.plot(self.train_accuracy_records, color="#1f77b4", label=f"training dataset: {round(float(self.train_accuracy_records[-1]), 3)}")
-            if self.val_inputs is not None and self.val_targets is not None:
-                plt.plot(self.val_accuracy_records, color="orange", label=f"validation dataset: {round(float(self.val_accuracy_records[-1]), 3)}")
-            plt.legend()
-            
-            figure.canvas.draw()
-            figure.canvas.flush_events()
-    
-    def fit(self, cost_fn, epochs = 1000, lr = 0.1, momentum = None, verbose = 1, live_stats = {"type": 0, "F1": 0, "F2": 1, "n_row": "limited"}):
         for n in [-1,0,1,2]:
-            if type(live_stats) == dict:
-                if live_stats["type"] == n:
-                    break
-                elif n == 2 and live_stats["type"] != n:
-                    raise ValueError("unknown value for live_stats['type']")
-            elif type(live_stats) == int:
-                if live_stats == n:
-                    break
-                elif n == 2 and live_stats != n:
-                    raise ValueError("unknown value for live_stats")
-                    
-        if type(live_stats) == dict:
-            if live_stats["n_row"] != "full" and live_stats["n_row"] != "limited":
-                raise ValueError("unknown value for live_stats['n_row']")
+            if live_stats.type == n:
+                break
+            elif n == 2 and live_stats.type != n:
+                raise ValueError("unknown value for live_stats['type']")
+
+        if live_stats.row_select != "full" and live_stats.row_select != "limited":
+            raise ValueError("unknown value for live_stats.row_select")
                 
         self.__is_trained__ = True
         import copy
@@ -433,17 +438,14 @@ class Sequential():
                 if cost_fn == name:
                     self.val_cost_fn = function
                     break
-        
-        if type(live_stats) == int:
-            live_stats = {"type": live_stats, "F1": 0, "F2": 1, "n_row": "limited"}
-        
-        if live_stats["type"] == -1:
+
+        if live_stats.type == -1:
             update = 50
             figure = plt.figure()
-        elif live_stats["type"] == 0:
+        elif live_stats.type == 0:
             update = 1
             figure = None
-        elif live_stats["type"] == 1 or live_stats["type"] == 2:
+        elif live_stats.type == 1 or live_stats.type == 2: 
             update = 25
             figure = plt.figure()
             
@@ -459,9 +461,9 @@ class Sequential():
                         self.val_accuracy_records.append(self.accuracy(self.val_targets, self.val_outputs))
                 self.backward(lr, momentum)
                 
-                if live_stats != 0:
+                if live_stats.type != 0:
                     if epoch % update == 0:
-                        self.live_stats(ls_type = live_stats["type"], figure = figure, F1 = live_stats["F1"], F2 = live_stats["F2"], n_row = live_stats["n_row"])
+                        live_stats.run(self, figure)
                             
                 self.verbose(verbose, epoch, epochs, start)
         else:
@@ -476,9 +478,9 @@ class Sequential():
                         self.val_accuracy_records.append(self.accuracy(self.val_targets, self.val_outputs))
                 self.backward(lr, momentum)
                 
-                if live_stats != 0:
+                if live_stats.type != 0:
                     if epoch % update == 0:
-                        self.live_stats(ls_type = live_stats["type"], figure = figure, F1 = live_stats["F1"], F2 = live_stats["F2"], n_row = live_stats["n_row"])
+                        live_stats.run(self, figure)
                         
                 self.verbose(verbose, epoch, epochs, start)
         
